#include <stdio.h>
#include <stdlib.h>
#include "align.h"

#ifndef SIMPLE_EST_TABLE
#ifndef EST_TABLE_UNIT_WIDTH
#define EST_TABLE_UNIT_WIDTH 16
#endif
#endif

#ifdef SIMPLE_EST_TABLE
#define put_estimate(estimator, i, j, estimate) \
  set_estimate(estimator, i, j, estimate)
#define get_estimate(estimator, i, j) \
  lookup_estimate(estimator, i, j)
#else
#define put_estimate(estimator, i, j, estimate) \
  ((estimator)->table[(estimator)->col_al_p->seq[0].len \
		      - (i) + (j)][((i) < (j)) ? (i) : (j)] = (estimate))
#define get_estimate(estimator, i, j) \
  ((estimator)->table[(estimator)->col_al_p->seq[0].len \
		      - (i) + (j)][((i) < (j)) ? (i) : (j)])
#endif

void estimator_setup(estimator_t estimator)
{
  int i, j, k, row, col, edge_cost, min_cost, new_cost, is_visited;

  if(estimator->dir == FORWARD){
    put_estimate(estimator, estimator->col_al_p->seq[0].len,
		 estimator->row_al_p->seq[0].len, 0);
    for(col = estimator->col_al_p->seq[0].len; col >= 0; col--){
      for(row = estimator->row_al_p->seq[0].len; row >= 0; row--){
	is_visited = 0;
	if((col != estimator->col_al_p->seq[0].len)
	   && (row != estimator->row_al_p->seq[0].len)){
	  is_visited = 1;
	  edge_cost = 0;
	  for(i = 0; i < estimator->col_al_p->dim; i++){
	    for(j = 0; j < estimator->row_al_p->dim; j++){
	      edge_cost += matching_cost(estimator->col_al_p->seq[i].str[col],
					 estimator->row_al_p->seq[j].str[row]);
	    }
	  }
	  min_cost = get_estimate(estimator, col + 1, row + 1) + edge_cost;
	}
	if(col != estimator->col_al_p->seq[0].len){
	  k = 0; 
	  for(i = 0; i < estimator->col_al_p->dim; i++){
	    if(is_normal_char(estimator->col_al_p->seq[i].str[col])){
	      k++;
	    }
	  }
	  edge_cost = 0;
	  for(j = 0; j < estimator->row_al_p->dim; j++){
	    if(((row == 0)
		|| (estimator->row_al_p->seq[j].str[row - 1]
		    == EXTERNAL_GAP_CHAR))
	       || ((row == estimator->row_al_p->seq[0].len)
		   || (estimator->row_al_p->seq[j].str[row]
		       == EXTERNAL_GAP_CHAR))){
	      edge_cost += k * (-external_gap_penalty);
	    }else{
	      edge_cost += k * (-internal_gap_penalty);
	    }
	  }
	  new_cost = get_estimate(estimator, col + 1, row) + edge_cost;
	  if((!is_visited) || (new_cost < min_cost)){
	    is_visited = 1;
	    min_cost = new_cost;
	  }
	}
	if(row != estimator->row_al_p->seq[0].len){
	  k = 0; 
	  for(i = 0; i < estimator->row_al_p->dim; i++){
	    if(is_normal_char(estimator->row_al_p->seq[i].str[row])){
	      k++;
	    }
	  }
	  edge_cost = 0;
	  for(j = 0; j < estimator->col_al_p->dim; j++){
	    if(((col == 0)
		|| (estimator->col_al_p->seq[j].str[col - 1]
		    == EXTERNAL_GAP_CHAR))
	       || ((col == estimator->col_al_p->seq[0].len)
		   || (estimator->col_al_p->seq[j].str[col]
		       == EXTERNAL_GAP_CHAR))){
	      edge_cost += k * (-external_gap_penalty);
	    }else{
	      edge_cost += k * (-internal_gap_penalty);
	    }
	  }
	  new_cost = get_estimate(estimator, col, row + 1) + edge_cost;
	  if((!is_visited) || (new_cost < min_cost)){
	    is_visited = 1;
	    min_cost = new_cost;
	  }
	}
	if(is_visited){
	  put_estimate(estimator, col, row, min_cost);
	}
      }
    }
  }
  if(estimator->dir == BACKWARD){
    put_estimate(estimator, 0, 0, 0);
    for(col = 0; col < estimator->col_al_p->seq[0].len + 1; col++){
      for(row = 0; row < estimator->row_al_p->seq[0].len + 1; row++){
	is_visited = 0;
	if((col != 0) && (row != 0)){
	  is_visited = 1;
	  edge_cost = 0;
	  for(i = 0; i < estimator->col_al_p->dim; i++){
	    for(j = 0; j < estimator->row_al_p->dim; j++){
	      edge_cost
		+= matching_cost(estimator->col_al_p->seq[i].str[col - 1],
				 estimator->row_al_p->seq[j].str[row - 1]);
	    }
	  }
	  min_cost = get_estimate(estimator, col - 1, row - 1) + edge_cost;
	}
	if(col != 0){
	  k = 0; 
	  for(i = 0; i < estimator->col_al_p->dim; i++){
	    if(is_normal_char(estimator->col_al_p->seq[i].str[col - 1])){
	      k++;
	    }
	  }
	  edge_cost = 0;
	  for(j = 0; j < estimator->row_al_p->dim; j++){
	    if(((row == 0)
		|| (estimator->row_al_p->seq[j].str[row - 1]
		    == EXTERNAL_GAP_CHAR))
	       || ((row == estimator->row_al_p->seq[0].len)
		   || (estimator->row_al_p->seq[j].str[row]
		       == EXTERNAL_GAP_CHAR))){
	      edge_cost += k * (-external_gap_penalty);
	    }else{
	      edge_cost += k * (-internal_gap_penalty);
	    }
	  }
	  new_cost = get_estimate(estimator, col - 1, row) + edge_cost;
	  if((!is_visited) || (new_cost < min_cost)){
	    is_visited = 1;
	    min_cost = new_cost;
	  }
	}
	if(row != 0){
	  k = 0; 
	  for(i = 0; i < estimator->row_al_p->dim; i++){
	    if(is_normal_char(estimator->row_al_p->seq[i].str[row - 1])){
	      k++;
	    }
	  }
	  edge_cost = 0;
	  for(j = 0; j < estimator->col_al_p->dim; j++){
	    if(((col == 0)
		|| (estimator->col_al_p->seq[j].str[col - 1]
		    == EXTERNAL_GAP_CHAR))
	       || ((col == estimator->col_al_p->seq[0].len)
		   || (estimator->col_al_p->seq[j].str[col]
		       == EXTERNAL_GAP_CHAR))){
	      edge_cost += k * (-external_gap_penalty);
	    }else{
	      edge_cost += k * (-internal_gap_penalty);
	    }
	  }
	  new_cost = get_estimate(estimator, col, row - 1) + edge_cost;
	  if((!is_visited) || (new_cost < min_cost)){
	    is_visited = 1;
	    min_cost = new_cost;
	  }
	}
	if(is_visited){
	  put_estimate(estimator, col, row, min_cost);
	}
      }
    }
  }
}

estimator_t estimator_init(alignment_t *al_p0, alignment_t *al_p1, int dir)
{
  estimator_t t;
#ifndef SIMPLE_EST_TABLE
  int i, n, width, table_size, temp_table_size, *temp_table;
#endif

  if((t = (estimator_t)malloc(sizeof(struct estimator))) == NULL){
    fprintf(stderr, "Can't allocate memory for estimator table\n");
    exit(-1);
  }
  t->dir = dir;
  t->col_al_p = al_p0;
  t->row_al_p = al_p1;
#ifdef SIMPLE_EST_TABLE
  if((t->table = (int *)malloc(sizeof(int) * (al_p0->seq[0].len + 1)
			       * (al_p1->seq[0].len + 1))) == NULL){
    fprintf(stderr, "Can't allocate memory for estimate table\n");
    exit(-1);
  }
#else
  if((t->table
      = (int **)malloc(sizeof(int *) * (al_p0->seq[0].len
					+ al_p1->seq[0].len + 1))) == NULL){
    fprintf(stderr, "Can't allocate memory for estimate table\n");
    exit(-1);
  }
  width = EST_TABLE_UNIT_WIDTH;
  if(al_p0->seq[0].len < al_p1->seq[0].len){
    if(width > al_p0->seq[0].len){
      width = al_p0->seq[0].len;
    }
    table_size
      = (al_p1->seq[0].len - al_p0->seq[0].len + 1) * (al_p0->seq[0].len + 1)
	+ 2 * al_p0->seq[0].len * width - width * (width - 1);
  }else{
    if(width > al_p1->seq[0].len){
      width = al_p1->seq[0].len;
    }
    table_size
      = (al_p0->seq[0].len - al_p1->seq[0].len + 1) * (al_p1->seq[0].len + 1)
	+ 2 * al_p1->seq[0].len * width - width * (width - 1);
  }
  if((temp_table_size
      = (al_p0->seq[0].len + 1) * (al_p1->seq[0].len + 1) - table_size) > 0){
    if((temp_table = (int *)malloc(sizeof(int) * temp_table_size)) == NULL){
      fprintf(stderr, "Can't allocate memory for estimate table\n");
      exit(-1);
    }
  }
  if(al_p0->seq[0].len < al_p1->seq[0].len){
    if((t->table[al_p0->seq[0].len]
	= (int *)malloc(sizeof(int) * table_size)) == NULL){
      fprintf(stderr, "Can't allocate memory for estimate table\n");
      exit(-1);
    }
    n = al_p0->seq[0].len + 1;
    for(i = al_p0->seq[0].len + 1; i < al_p1->seq[0].len + 1; i++){
      t->table[i] = t->table[al_p0->seq[0].len] + n;
      n += al_p0->seq[0].len + 1;
    }
    for(i = 1; i < width + 1; i++){
      t->table[al_p0->seq[0].len - i] = t->table[al_p0->seq[0].len] + n;
      n += al_p0->seq[0].len + 1 - i;
      t->table[al_p1->seq[0].len + i] = t->table[al_p0->seq[0].len] + n;
      n += al_p0->seq[0].len + 1 - i;
    }
    n = 0;
    for(i = width + 1; i < al_p0->seq[0].len + 1; i++){
      t->table[al_p0->seq[0].len - i] = temp_table + n;
      n += al_p0->seq[0].len + 1 - i;
      t->table[al_p1->seq[0].len + i] = temp_table + n;
      n += al_p0->seq[0].len + 1 - i;
    }
  }else{
    if((t->table[al_p1->seq[0].len]
	= (int *)malloc(sizeof(int) * table_size)) == NULL){
      fprintf(stderr, "Can't allocate memory for estimate table\n");
      exit(-1);
    }
    n = al_p1->seq[0].len + 1;
    for(i = al_p1->seq[0].len + 1; i < al_p0->seq[0].len + 1; i++){
      t->table[i] = t->table[al_p1->seq[0].len] + n;
      n += al_p1->seq[0].len + 1;
    }
    for(i = 1; i < width + 1; i++){
      t->table[al_p1->seq[0].len - i] = t->table[al_p1->seq[0].len] + n;
      n += al_p1->seq[0].len + 1 - i;
      t->table[al_p0->seq[0].len + i] = t->table[al_p1->seq[0].len] + n;
      n += al_p1->seq[0].len + 1 - i;
    }
    n = 0;
    for(i = width + 1; i < al_p1->seq[0].len + 1; i++){
      t->table[al_p1->seq[0].len - i] = temp_table + n;
      n += al_p1->seq[0].len + 1 - i;
      t->table[al_p0->seq[0].len + i] = temp_table + n;
      n += al_p1->seq[0].len + 1 - i;
    }
  }
#endif
  estimator_setup(t);
#ifndef SIMPLE_EST_TABLE
  if(temp_table_size > 0){
    free(temp_table);
  }
  if(al_p0->seq[0].len < al_p1->seq[0].len){
    t->start = al_p0->seq[0].len - width;
    t->end = al_p1->seq[0].len + width;
  }else{
    t->start = al_p1->seq[0].len - width;
    t->end = al_p0->seq[0].len + width;
  }
#endif
  return t;
}

void estimator_finish(estimator_t estimator)
{
#ifndef SIMPLE_EST_TABLE
  if(estimator->col_al_p->seq[0].len < estimator->row_al_p->seq[0].len){
    free(estimator->table[estimator->col_al_p->seq[0].len]);
  }else{
    free(estimator->table[estimator->row_al_p->seq[0].len]);
  }
#endif
  free(estimator->table);
  free(estimator);
}

#ifndef SIMPLE_EST_TABLE
void estimator_extend(estimator_t estimator, int i, int j)
{
  int n, new_width, table_size, temp_table_size, *temp_table;

  if(estimator->col_al_p->seq[0].len - i + j < estimator->start){
    new_width = (i - j - 1) / EST_TABLE_UNIT_WIDTH + 1;
  }
  if(estimator->col_al_p->seq[0].len - i + j > estimator->end){
    new_width = (j - i - 1) / EST_TABLE_UNIT_WIDTH + 1;
  }
  new_width *= EST_TABLE_UNIT_WIDTH;
  if(estimator->col_al_p->seq[0].len < estimator->row_al_p->seq[0].len){
    if(new_width > estimator->col_al_p->seq[0].len){
      new_width = estimator->col_al_p->seq[0].len;
    }
    table_size
      = (estimator->row_al_p->seq[0].len
	 - estimator->col_al_p->seq[0].len + 1)
	* (estimator->col_al_p->seq[0].len + 1)
	  + 2 * estimator->col_al_p->seq[0].len * new_width
	    - new_width * (new_width - 1);
  }else{
    if(new_width > estimator->row_al_p->seq[0].len){
      new_width = estimator->row_al_p->seq[0].len;
    }
    table_size
      = (estimator->col_al_p->seq[0].len
	 - estimator->row_al_p->seq[0].len + 1)
	* (estimator->row_al_p->seq[0].len + 1)
	  + 2 * estimator->row_al_p->seq[0].len * new_width
	    - new_width * (new_width - 1);
  }
  if((temp_table_size = (estimator->col_al_p->seq[0].len + 1)
      * (estimator->row_al_p->seq[0].len + 1) - table_size) > 0){
    if((temp_table = (int *)malloc(sizeof(int) * temp_table_size)) == NULL){
      fprintf(stderr, "Can't allocate memory for estimate table\n");
      exit(-1);
    }
  }
  if(estimator->col_al_p->seq[0].len < estimator->row_al_p->seq[0].len){
    free(estimator->table[estimator->col_al_p->seq[0].len]);
    if((estimator->table[estimator->col_al_p->seq[0].len]
	= (int *)malloc(sizeof(int) * table_size)) == NULL){
      fprintf(stderr, "Can't allocate memory for estimate table\n");
      exit(-1);
    }
    n = estimator->col_al_p->seq[0].len + 1;
    for(i = estimator->col_al_p->seq[0].len + 1;
	i < estimator->row_al_p->seq[0].len + 1; i++){
      estimator->table[i]
	= estimator->table[estimator->col_al_p->seq[0].len] + n;
      n += estimator->col_al_p->seq[0].len + 1;
    }
    for(i = 1; i < new_width + 1; i++){
      estimator->table[estimator->col_al_p->seq[0].len - i]
	= estimator->table[estimator->col_al_p->seq[0].len] + n;
      n += estimator->col_al_p->seq[0].len + 1 - i;
      estimator->table[estimator->row_al_p->seq[0].len + i]
	= estimator->table[estimator->col_al_p->seq[0].len] + n;
      n += estimator->col_al_p->seq[0].len + 1 - i;
    }
    n = 0;
    for(i = new_width + 1; i < estimator->col_al_p->seq[0].len + 1; i++){
      estimator->table[estimator->col_al_p->seq[0].len - i] = temp_table + n;
      n += estimator->col_al_p->seq[0].len + 1 - i;
      estimator->table[estimator->row_al_p->seq[0].len + i] = temp_table + n;
      n += estimator->col_al_p->seq[0].len + 1 - i;
    }
  }else{
    free(estimator->table[estimator->row_al_p->seq[0].len]);
    if((estimator->table[estimator->row_al_p->seq[0].len]
	= (int *)malloc(sizeof(int) * table_size)) == NULL){
      fprintf(stderr, "Can't allocate memory for estimate table\n");
      exit(-1);
    }
    n = estimator->row_al_p->seq[0].len + 1;
    for(i = estimator->row_al_p->seq[0].len + 1;
	i < estimator->col_al_p->seq[0].len + 1; i++){
      estimator->table[i]
	= estimator->table[estimator->row_al_p->seq[0].len] + n;
      n += estimator->row_al_p->seq[0].len + 1;
    }
    for(i = 1; i < new_width + 1; i++){
      estimator->table[estimator->row_al_p->seq[0].len - i]
	= estimator->table[estimator->row_al_p->seq[0].len] + n;
      n += estimator->row_al_p->seq[0].len + 1 - i;
      estimator->table[estimator->col_al_p->seq[0].len + i]
	= estimator->table[estimator->row_al_p->seq[0].len] + n;
      n += estimator->row_al_p->seq[0].len + 1 - i;
    }
    n = 0;
    for(i = new_width + 1; i < estimator->row_al_p->seq[0].len + 1; i++){
      estimator->table[estimator->row_al_p->seq[0].len - i] = temp_table + n;
      n += estimator->row_al_p->seq[0].len + 1 - i;
      estimator->table[estimator->col_al_p->seq[0].len + i] = temp_table + n;
      n += estimator->row_al_p->seq[0].len + 1 - i;
    }
  }
#ifdef VERBOSE
  fprintf(stderr, "Recalculating pairweise alignment .... ");
#endif
  estimator_setup(estimator);
#ifdef VERBOSE
  fprintf(stderr, "done\n");
#endif
  if(temp_table_size > 0){
    free(temp_table);
  }
#ifdef VERBOSE
  fprintf(stderr, "Estimate table extended: %d -> ",
	  estimator->end - estimator->start + 1);
#endif
  if(estimator->col_al_p->seq[0].len < estimator->row_al_p->seq[0].len){
    estimator->start = estimator->col_al_p->seq[0].len - new_width;
    estimator->end = estimator->row_al_p->seq[0].len + new_width;
  }else{
    estimator->start = estimator->row_al_p->seq[0].len - new_width;
    estimator->end = estimator->col_al_p->seq[0].len + new_width;
  }
#ifdef VERBOSE
  fprintf(stderr, "%d\n", estimator->end - estimator->start + 1);
#endif
}

void set_estimate(estimator_t estimator, int i, int j, int estimate)
{
  if(i > j){
    estimator->table[estimator->col_al_p->seq[0].len - i + j][j] = estimate;
  }else{
    estimator->table[estimator->col_al_p->seq[0].len - i + j][i] = estimate;
  }
}

int lookup_estimate(estimator_t estimator, int i, int j)
{
  int n;

  if(i > j){
    if((n = estimator->col_al_p->seq[0].len - i + j) < estimator->start){
      estimator_extend(estimator, i, j);
    }
    return estimator->table[n][j];
  }else{
    if((n = estimator->col_al_p->seq[0].len - i + j) > estimator->end){
      estimator_extend(estimator, i, j);
    }
    return estimator->table[n][i];
  }
}
#endif
