#include <stdio.h>
#include <stdlib.h>
#include "align.h"

int takeword(FILE *fp,char *buf)
{
  char c;
  int i;

  c = fgetc(fp);
  while ((c == ' ')||(c == '\t')||(c == '\n')||(c == '=')
	 ||(c == ',')) c = fgetc(fp);
  i = 0;
  while (1) {
    if (c == EOF) {buf[i] = '\0';return -1;}
    if ((c == ' ')||(c == '\t'))
      {buf[i] = '\0';return 0;}
    if (c == '\n') {buf[i] = '\0';return 1;}
    buf[i] = c;
    c = fgetc(fp);
    i++;
  }
}

void print_alignment(alignment_t *al_p, int line)
{
  int i, j, n;

  for(i = 0; i < al_p->seq[0].len; i += line){
    printf("\n");
    printf("  %-24d  %24d\n", i + 1, i + line);
    for(j = 0; j < al_p->dim; j++){
      printf("  ");
      for(n = 0; n < line; n++){
	if(i + n >= al_p->seq[0].len){
	  break;
	}
	if(al_p->seq[j].str[i + n] == INTERNAL_GAP_CHAR){
	  putchar('-');
	}else if(al_p->seq[j].str[i + n] == EXTERNAL_GAP_CHAR){
	  putchar('=');
	}else{
	  putchar(al_p->seq[j].str[i + n]);
	}
      }
      printf("\n");
    }
  }
}

void print_number_of_gaps(alignment_t *al_p, int *ig, int *eg, int *is, int *es,
			  int print_results)
{
  int i, j, k;
  int number_of_internal, number_of_external;
  int number_of_internal_gap_sequence, number_of_external_gap_sequence;
  int state; /* 0: char-char / -1: gap-char / 1: char-gap */

  if (al_p->dim < 2) return;

  number_of_internal = 0;
  number_of_external = 0;
  number_of_internal_gap_sequence = 0;
  number_of_external_gap_sequence = 0;

  for(i=1;i<al_p->dim; i++) {
    for(j=0;j<i;j++) {
      state = 0;
      for(k=0;k<al_p->seq[0].len;k++) {
	if(al_p->seq[i].str[k] < INTERNAL_GAP_CHAR) {
	  if(al_p->seq[j].str[k] < INTERNAL_GAP_CHAR) {
	    state = 0; 
	  } else if(al_p->seq[j].str[k] == INTERNAL_GAP_CHAR) {
	    number_of_internal++;
	    if (state <= 0) {
	      number_of_internal_gap_sequence++;
	      state = 1;
	    }
	  } else {
	    number_of_external++;
	    if (state <= 0) {
	      number_of_external_gap_sequence++;
	      state = 1;
	    }
	  }
	} else if(al_p->seq[i].str[k] == INTERNAL_GAP_CHAR)
	  {
	    if (al_p->seq[j].str[k] < INTERNAL_GAP_CHAR) {
	      number_of_internal++;
	      if (state >= 0) {
		number_of_internal_gap_sequence++;
		state = -1;
	      }
	    }
	  } else {
	    if(al_p->seq[j].str[k] < INTERNAL_GAP_CHAR) {
	      number_of_external++;
	      if (state >= 0) {
		number_of_external_gap_sequence++;
		state = -1;
	      }
	    }
	  }
      }
    }
  }

  (*ig) = number_of_internal;
  (*eg) = number_of_external;
  (*is) = number_of_internal_gap_sequence;
  (*es) = number_of_external_gap_sequence;

  if (print_results) {
    printf("Gap Information :\n");
    printf("  Number of gaps(indels) : %5d (%d+%d)\n", number_of_internal+number_of_external, number_of_internal, number_of_external);

/*  printf("  Number of internal gaps(indels) : %5d\n", number_of_internal);
    printf("  Number of external gaps(indels) : %5d\n\n", number_of_external);*/

    printf("  Number of gap sequences: %5d (%d+%d)\n", number_of_internal_gap_sequence+number_of_external_gap_sequence, number_of_internal_gap_sequence, number_of_external_gap_sequence);

/*  printf("  Number of internal gap sequences: %5d\n", number_of_internal_gap_sequence);
    printf("  Number of external gap sequences: %5d\n\n", number_of_external_gap_sequence);*/
  }
}


void print_subopt_answer(subopt_answer_t subanswer, int depth,
    node_t mid_node, node_t orig_node, node_t dest_node, int dir,
    int dim, alignment_t *alignment, int line, int print_subopt,
    solutions_p opt_sols, solutions2_p opt_sols2, solutions2_p opt_sols3)
{
  alignment_t *al_p;
  node_t node;
  int i, j, k, p, q, d, s;
  int ig, eg, tg, is, es;
  int slope;
  int al_len, al_dim;
  int status; /* on_best_path: 1 / otherwise: 0 */

  if((al_p = (alignment_t *)malloc(sizeof(alignment_t))) == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }

  al_dim = 0;
  for(i = 0; i < dim; i++){
    al_dim += alignment[i].dim;
  }
  al_p->dim = al_dim;

  al_len = 0;

  if (print_subopt) printf("\nDifferent regions: ");

  if (dir == FORWARD) {
    status = (is_on_the_best_path(mid_node, dir));
    d = 0;
    for(node = mid_node; node != orig_node; node = lookup_sub_fs_prev_node(node, subanswer, &d, depth)){
      al_len++;
      if (print_subopt) {
	if ((status) && (!(is_on_the_best_path(node, dir)))) {
	  status = 0;
	  printf("%d-", node->pos[0]);
	} else if ((!(status)) && (is_on_the_best_path(node, dir))) {
	  status = 1;
	  printf("%d ", node->pos[0]);
	}
      }
    }
  }

  if (dir == BACKWARD) {
    status = (is_on_the_best_path(mid_node, dir));
    d = 0;
    for(node = mid_node; node != dest_node; node = lookup_sub_bs_prev_node(node, subanswer, &d, depth)){
      al_len++;
      if (print_subopt) {
	if ((status) && (!(is_on_the_best_path(node, dir)))) {
	  status = 0;
	  printf("%d ", node->pos[0]);
	} else if ((!(status)) && (is_on_the_best_path(node, dir))) {
	  status = 1;
	  printf("%d ", node->pos[0]);
	}
      }
    }
  }

  if (print_subopt) {
    printf("\n");
    printf("Depth: %d\n", depth + 1);
  }

  if((al_p->seq = (sequence_t *)malloc(sizeof(sequence_t) * al_dim))
     == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }

/*  alloc_alignment(al_p, al_dim, al_len);*/

  k = 0;
  for(i = 0; i < dim; i++){
    for(j = 0; j < alignment[i].dim; j++){
      al_p->seq[k].len = al_len;
      if((al_p->seq[k].str
	  = (char *)malloc(sizeof(char) * (al_len + 1))) == NULL){
	fprintf(stderr, "Can't allocate memory for result alignment");
	exit(-1);
      }
      al_p->seq[k].str[al_len] = '\0';

      if(dir == FORWARD) {
	p = -1;
	d = 0;
	for(node = mid_node; node != orig_node;
	    node = lookup_sub_fs_prev_node(node, subanswer, &d, depth)){
	  p++;
	}
	q = mid_node->pos[i] - 1;
	d = 0;
	status = is_on_the_best_path(mid_node, dir);
	for(node = mid_node; node != orig_node;
	    node = lookup_sub_fs_prev_node(node, subanswer, &d, depth)){
	  if(node->pos[i] != lookup_sub_fs_prev_node_s(node, subanswer, &d, depth)->pos[i]){
	    if (is_on_the_best_path(node, dir)) {
	      if (status == 0) {
		al_p->seq[k].str[p--] = alignment[i].seq[j].str[q--];
		status = 1;    /*  DIFFERENT REGION  */
	      } else {
		al_p->seq[k].str[p--] = alignment[i].seq[j].str[q--];
	      }
	    } else {
	      status = 0;    /*  DIFFERENT REGION  */
	      al_p->seq[k].str[p--] = alignment[i].seq[j].str[q--];
	    }
	  }else{
	    status = is_on_the_best_path(node, dir);
	    if(((q + 1 > alignment[i].seq[0].len - 1)
		|| (alignment[i].seq[j].str[q + 1] == EXTERNAL_GAP_CHAR))
	       || ((q < 0)
		   || (alignment[i].seq[j].str[q] == EXTERNAL_GAP_CHAR))){
	      al_p->seq[k].str[p--] = EXTERNAL_GAP_CHAR;
	    }else{
	      al_p->seq[k].str[p--] = INTERNAL_GAP_CHAR;
	    }
	  }
	}
      }

      if (dir == BACKWARD) {
	p = al_len;
	for(node = mid_node; node != dest_node;
	    node = lookup_sub_bs_prev_node(node, subanswer, &d, depth)){
	  p--;
	}
	q = mid_node->pos[i];
	d = 0;
	status = is_on_the_best_path(mid_node, dir);
	for(node = mid_node; node != dest_node;
	    node = lookup_sub_bs_prev_node(node, subanswer, &d, depth)){
	  if(node->pos[i] != lookup_sub_bs_prev_node_s(node, subanswer, &d, depth)->pos[i]){
	    if (is_on_the_best_path(node, dir)) {
	      if (status == 0) {
		al_p->seq[k].str[p++] = alignment[i].seq[j].str[q++];
		status = 1;    /*  DIFFERENT REGION  */
	      } else {
		al_p->seq[k].str[p++] = alignment[i].seq[j].str[q++];
	      }
	    } else {
	      status = 0;    /*  DIFFERENT REGION  */
	      al_p->seq[k].str[p++] = alignment[i].seq[j].str[q++];
	    }
	  }else{
	    status = is_on_the_best_path(node, dir);
	    if(((q > alignment[i].seq[0].len - 1)
		|| (alignment[i].seq[j].str[q] == EXTERNAL_GAP_CHAR))
	       || ((q - 1 < 0)
		   || (alignment[i].seq[j].str[q - 1] == EXTERNAL_GAP_CHAR))){
	      al_p->seq[k].str[p++] = EXTERNAL_GAP_CHAR;
	    }else{
	      al_p->seq[k].str[p++] = INTERNAL_GAP_CHAR;
	    }
	  }
	}
      }

      k++;
    }
  }
  if (print_subopt) {
    printf("Suboptimal alignment:\n");
    print_alignment(al_p, line);
    printf("\n");
    print_number_of_gaps(al_p, (&ig), (&eg), (&is), (&es), 1);
    printf("\n------\n");
  }

  if (opt_sols2 != NULL) {
    (opt_sols2->number_of_sols)++;
    slope = weighted_score(al_p, weights3);
    if (opt_sols2->max_alignment == NULL) {
      opt_sols2->max_alignment = al_p;
      opt_sols2->min_alignment = al_p;
      opt_sols2->max_slope = slope;
      opt_sols2->min_slope = slope;
      opt_sols2->number_of_max_sols = 1;
      opt_sols2->number_of_min_sols = 1;
      if (print_subopt) {
	printf("This is temporally both min and max\n======\n");
      }
    } else {
      s = 1;
      if (opt_sols2->max_slope == slope) {
	(opt_sols2->number_of_max_sols)++;
	if (print_subopt) {
	  printf("This is temporally max\n======\n");
	}
      } else if (opt_sols2->max_slope < slope) {
	opt_sols2->max_slope = slope;
	opt_sols2->number_of_max_sols = 1;
	if (opt_sols2->max_alignment != opt_sols2->min_alignment) {
	  free_alignment(opt_sols2->max_alignment, 1);
	}
	s = 0;
	opt_sols2->max_alignment = al_p;
      }

      if (opt_sols2->min_slope == slope) {
	(opt_sols2->number_of_min_sols)++;
	if (print_subopt) {
	  printf("This is temporally min\n======\n");
	}
      } else if (opt_sols2->min_slope > slope) {
	opt_sols2->min_slope = slope;
	opt_sols2->number_of_min_sols = 1;
	if (opt_sols2->max_alignment != opt_sols2->min_alignment) {
	  free_alignment(opt_sols2->min_alignment, 1);
	}
	s = 0;
	opt_sols2->min_alignment = al_p;
      }
      if (s) free_alignment(al_p, 1);
    }
  } else if (opt_sols3 != NULL) {
    (opt_sols3->number_of_sols)++;
    slope = score_slope(al_p);
    if (opt_sols3->max_alignment == NULL) {
      opt_sols3->max_alignment = al_p;
      opt_sols3->min_alignment = al_p;
      opt_sols3->max_slope = slope;
      opt_sols3->min_slope = slope;
      opt_sols3->number_of_max_sols = 1;
      opt_sols3->number_of_min_sols = 1;
      if (print_subopt) {
	printf("This is temporally both min and max\n======\n");
      }
    } else {
      s = 1;
      if (opt_sols3->max_slope == slope) {
	(opt_sols3->number_of_max_sols)++;
	if (print_subopt) {
	  printf("This is temporally max\n======\n");
	}
      } else if (opt_sols3->max_slope < slope) {
	opt_sols3->max_slope = slope;
	opt_sols3->number_of_max_sols = 1;
	if (opt_sols3->max_alignment != opt_sols3->min_alignment) {
	  free_alignment(opt_sols3->max_alignment, 1);
	}
	s = 0;
	opt_sols3->max_alignment = al_p;
      }

      if (opt_sols3->min_slope == slope) {
	(opt_sols3->number_of_min_sols)++;
	if (print_subopt) {
	  printf("This is temporally min\n======\n");
	}
      } else if (opt_sols3->min_slope > slope) {
	opt_sols3->min_slope = slope;
	opt_sols3->number_of_min_sols = 1;
	if (opt_sols3->max_alignment != opt_sols3->min_alignment) {
	  free_alignment(opt_sols3->min_alignment, 1);
	}
	s = 0;
	opt_sols3->min_alignment = al_p;
      }
      if (s) free_alignment(al_p, 1);
    }
  } else if (opt_sols != NULL) {
    print_number_of_gaps(al_p, (&ig), (&eg), (&is), (&es), 0);
    tg = ig + eg;
    (opt_sols->number_of_sols)++;
    if (opt_sols->max_alignment == NULL) {
      opt_sols->max_alignment = al_p;
      opt_sols->min_alignment = al_p;
      opt_sols->max_gap = tg;
      opt_sols->min_gap = tg;
      opt_sols->number_of_max_sols = 1;
      opt_sols->number_of_min_sols = 1;
      if (print_subopt) {
	printf("This is temporally both min and max\n======\n");
      }
    } else {
      s = 1;
      if (opt_sols->max_gap == tg) {
	(opt_sols->number_of_max_sols)++;
	if (print_subopt) {
	  printf("This is temporally max\n======\n");
	}
      } else if (opt_sols->max_gap < tg) {
	opt_sols->max_gap = tg;
	opt_sols->number_of_max_sols = 1;
	if (opt_sols->max_alignment != opt_sols->min_alignment) {
	  free_alignment(opt_sols->max_alignment, 1);
	}
	s = 0;
	opt_sols->max_alignment = al_p;
      }

      if (opt_sols->min_gap == tg) {
	(opt_sols->number_of_min_sols)++;
	if (print_subopt) {
	  printf("This is temporally min\n======\n");
	}
      } else if (opt_sols->min_gap > tg) {
	opt_sols->min_gap = tg;
	opt_sols->number_of_min_sols = 1;
	if (opt_sols->max_alignment != opt_sols->min_alignment) {
	  free_alignment(opt_sols->min_alignment, 1);
	}
	s = 0;
	opt_sols->min_alignment = al_p;
      }
      if (s) free_alignment(al_p, 1);
    }
  } else {
    free_alignment(al_p, 1);
    return;
  }
}

node_t lookup_sub_fs_prev_node(node_t node,
    subopt_answer_t subanswer, int *dp, int depth)
{
  if (((*dp) <= depth) && (node == lookup_f_of_subopt_answer(subanswer, *dp))) {
    (*dp)++;
    return lookup_t_of_subopt_answer(subanswer, (*dp-1));
  } else return lookup_fs_prev_node(node);
}

node_t lookup_sub_bs_prev_node(node_t node,
    subopt_answer_t subanswer, int *dp, int depth)
{
  if (((*dp) <= depth) && (node == lookup_f_of_subopt_answer(subanswer, *dp))) {
    (*dp)++;
    return lookup_t_of_subopt_answer(subanswer, (*dp-1));
  } else return lookup_bs_prev_node(node);
}

node_t lookup_sub_fs_prev_node_s(node_t node,
    subopt_answer_t subanswer, int *dp, int depth)
{
  if (((*dp) <= depth) && (node == lookup_f_of_subopt_answer(subanswer, *dp))) {
    return lookup_t_of_subopt_answer(subanswer, *dp);
  } else return lookup_fs_prev_node(node);
}

node_t lookup_sub_bs_prev_node_s(node_t node,
    subopt_answer_t subanswer, int *dp, int depth)
{
  if (((*dp) <= depth) && (node == lookup_f_of_subopt_answer(subanswer, *dp))) {
    return lookup_t_of_subopt_answer(subanswer, *dp);
  } else return lookup_bs_prev_node(node);
}

void alloc_alignment(alignment_t *al_p, int dim, int al_len)
{
  int i;

  if((al_p = (alignment_t *)malloc(sizeof(alignment_t))) == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  al_p->dim = dim;
  if((al_p->seq = (sequence_t *)malloc(sizeof(sequence_t) * dim))
     == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  for(i=0; i<dim; i++) {
    al_p->seq[i].len = al_len;
    if((al_p->seq[i].str
	  = (char *)malloc(sizeof(char) * (al_len + 1))) == NULL){
	fprintf(stderr, "Can't allocate memory for result alignment");
	exit(-1);
      }
    al_p->seq[i].str[al_len] = '\0';
  }
}

void free_alignment(alignment_t *al_p, int groups)
{
  int i, j;

  for(j=0; j<groups; j++) {
    for(i=0; i<al_p[j].dim; i++) {
      free(al_p[j].seq[i].str);
    }
    free(al_p[j].seq);
  }

  free(al_p);
}

void print_boundaries(solutions_p boundaries, int line, int size)
{
  int i;
  int ig, eg, is, es;

  printf("Done %d iterations.\n", size);

  for(i=0;i<size;i++) {
    if (boundaries[i].max_gap != boundaries[i].min_gap) {
      printf("-\n%d: between %d and %d  <boundary>\n",
	     i, boundaries[i].left, boundaries[i].right);
      print_alignment(boundaries[i].max_alignment, line);
      print_number_of_gaps(boundaries[i].max_alignment, &ig, &eg, &is, &es, 1);
      print_alignment(boundaries[i].min_alignment, line);
      print_number_of_gaps(boundaries[i].min_alignment, &ig, &eg, &is, &es, 1);
    } else {
      printf("%d: between %d and %d  <Not boundary>\n", i, boundaries[i].left, boundaries[i].right);
      print_alignment(boundaries[i].max_alignment, line);
      print_number_of_gaps(boundaries[i].max_alignment, &ig, &eg, &is, &es, 1);
    }
  }

  for(i=0;i<size;i++) {
    if (boundaries[i].max_gap != boundaries[i].min_gap) {
      printf("==\n%d: between %d and %d  <boundary>\n",
	     i, boundaries[i].left, boundaries[i].right);
      printf("  Gap Penalty: %d/%d   Score: %d\n  #Gap: %d-%d",
	     boundaries[i].gap_penalty, boundaries[i].gap_denominator,
	     boundaries[i].score, boundaries[i].min_gap, boundaries[i].max_gap);
      printf("  #Solutions: %d (%d/%d)\n",
	     boundaries[i].number_of_sols, boundaries[i].number_of_max_sols,
	     boundaries[i].number_of_min_sols);
    } else {
      printf("%d: between %d and %d  <Not boundary>\n", i, boundaries[i].left, boundaries[i].right);
      printf("  Gap Penalty: %d/%d   Score: %d  #Gap: %d\n",
	     boundaries[i].gap_penalty, boundaries[i].gap_denominator,
	     boundaries[i].score, boundaries[i].max_gap);
      printf("  #Solutions: %d\n", boundaries[i].number_of_sols);
    }
  }
}

void print_boundaries2(solutions2_p boundaries, int line, int size, int dim)
{
  int i,j,k;
  int ig, eg, is, es;

  printf("Done %d iterations.\n", size);

  for(i=0;i<size;i++) {
    if (boundaries[i].max_slope != boundaries[i].min_slope) {
      printf("-\n%d: between %d and %d  <boundary>\n",
	     i, boundaries[i].left, boundaries[i].right);
      print_alignment(boundaries[i].max_alignment, line);
      print_number_of_gaps(boundaries[i].max_alignment, &ig, &eg, &is, &es, 1);
      print_alignment(boundaries[i].min_alignment, line);
      print_number_of_gaps(boundaries[i].min_alignment, &ig, &eg, &is, &es, 1);
    } else {
      printf("%d: between %d and %d  <Not boundary>\n", i, boundaries[i].left, boundaries[i].right);
      print_alignment(boundaries[i].max_alignment, line);
      print_number_of_gaps(boundaries[i].max_alignment, &ig, &eg, &is, &es, 1);
    }
  }

  for(i=0;i<size;i++) {
    if (boundaries[i].max_slope != boundaries[i].min_slope) {
      printf("==\n%d: between %d and %d  <boundary>\n",
	     i, boundaries[i].left, boundaries[i].right);
      printf("p =%3d, q=%3d  (p+q=%3d)\n",
	     boundaries[i].p, boundaries[i].q,
	     boundaries[i].p + boundaries[i].q);
      printf("Score = %d / max(%d, %d) min(%d, %d)\n",
	     boundaries[i].score,
	     weighted_score(boundaries[i].max_alignment,weights1),
	     weighted_score(boundaries[i].max_alignment,weights2),
	     weighted_score(boundaries[i].min_alignment,weights1),
	     weighted_score(boundaries[i].min_alignment,weights2));
      printf("weight matrix:\n");
      for(j=0;j<dim;j++) for(k=0;k<dim;k++)
	weights[j][k] = boundaries[i].q * weights1[j][k]
	              + boundaries[i].p * weights2[j][k];
      normalize_weights(weights, dim);
      print_weights(weights, dim);
      printf("  #Solutions: %d (%d/%d)\n",
	     boundaries[i].number_of_sols, boundaries[i].number_of_max_sols,
	     boundaries[i].number_of_min_sols);
    } else {
      printf("%d: between %d and %d  <Not boundary>\n", i, boundaries[i].left, boundaries[i].right);
      printf("p =%3d, q=%3d  (p+q=%3d)\n",
	     boundaries[i].p, boundaries[i].q,
	     boundaries[i].p + boundaries[i].q);
      printf("Score = %d / all(%d, %d)\n",
	     boundaries[i].score,
	     weighted_score(boundaries[i].max_alignment,weights1),
	     weighted_score(boundaries[i].max_alignment,weights2));
      for(j=0;j<dim;j++) for(k=0;k<dim;k++)
	weights[j][k] = boundaries[i].q * weights1[j][k]
	              + boundaries[i].p * weights2[j][k];
      normalize_weights(weights, dim);
      printf("  #Solutions: %d\n",
	     boundaries[i].number_of_sols);
    }
  }
}

void print_boundaries3(solutions2_p boundaries, int line, int size, int dim)
{
  int i,j,k;
  int ig, eg, is, es;

  printf("Done %d iterations.\n", size);

  for(i=0;i<size;i++) {
    if (boundaries[i].max_slope != boundaries[i].min_slope) {
      printf("-\n%d: between %d and %d  <boundary>\n",
	     i, boundaries[i].left, boundaries[i].right);
      print_alignment(boundaries[i].max_alignment, line);
      print_number_of_gaps(boundaries[i].max_alignment, &ig, &eg, &is, &es, 1);
      print_alignment(boundaries[i].min_alignment, line);
      print_number_of_gaps(boundaries[i].min_alignment, &ig, &eg, &is, &es, 1);
    } else {
      printf("%d: between %d and %d  <Not boundary>\n", i, boundaries[i].left, boundaries[i].right);
      print_alignment(boundaries[i].max_alignment, line);
      print_number_of_gaps(boundaries[i].max_alignment, &ig, &eg, &is, &es, 1);
    }
  }

  for(i=0;i<size;i++) {
    if (boundaries[i].max_slope != boundaries[i].min_slope) {
      printf("==\n%d: between %d and %d  <boundary>\n",
	     i, boundaries[i].left, boundaries[i].right);
      printf("p =%3d, q=%3d  (p+q=%3d)\n",
	     boundaries[i].p, boundaries[i].q,
	     boundaries[i].p + boundaries[i].q);
      printf("Score = %d / max(%d, %d) min(%d, %d)\n",
	     boundaries[i].score,
	     another_score1(boundaries[i].max_alignment),
	     another_score2(boundaries[i].max_alignment),
	     another_score1(boundaries[i].min_alignment),
	     another_score2(boundaries[i].min_alignment));
      printf("  #Solutions: %d (%d/%d)\n",
	     boundaries[i].number_of_sols, boundaries[i].number_of_max_sols,
	     boundaries[i].number_of_min_sols);
    } else {
      printf("%d: between %d and %d  <Not boundary>\n", i, boundaries[i].left, boundaries[i].right);
      printf("p =%3d, q=%3d  (p+q=%3d)\n",
	     boundaries[i].p, boundaries[i].q,
	     boundaries[i].p + boundaries[i].q);
      printf("Score = %d / all(%d, %d)\n",
	     boundaries[i].score,
	     another_score1(boundaries[i].max_alignment),
	     another_score2(boundaries[i].max_alignment));
      printf("  #Solutions: %d\n",
	     boundaries[i].number_of_sols);
    }
  }
}

void read_weight_file(FILE *w_fp, int **ws, int dim)
{
  int i, j, newline;
  char buf[BUFSIZE];

  for(i=0;i<dim;i++) {
    for(j=0;j<dim;j++) {
      newline = takeword(w_fp, buf);
      sscanf(buf,"%d",&(ws[i][j]));
    }
    while (!newline) {newline = takeword(w_fp, buf);}
  }
}

/*
void read_weight_file_original(FILE *w_fp, int **ws, int dim)
{
  int i, j;

  for(i=0;i<dim;i++) {
    for(j=0;j<dim;j++) {
      fscanf(w_fp, "%d", &(ws[i][j]));
    }
  }
}
*/

void print_weights(int **ws, int dim)
{
  int i,j;

  for(i=0;i<dim;i++) {
    for(j=0;j<dim;j++) {
      printf(" %3d", ws[i][j]);
    }
    printf("\n");
  }
}
